@@ -2,8 +2,8 @@
 /*
  This file is part of SLRN.
 
- Copyright (c) 1994, 1999 John E. Davis <davis@space.mit.edu>
- Copyright (c) 2001-2003 Thomas Schultz <tststs@gmx.de>
+ Copyright (c) 1994, 1999, 2007-2008 John E. Davis <jed@jedsoft.org>
+ Copyright (c) 2001-2006 Thomas Schultz <tststs@gmx.de>
 
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
@@ -63,6 +63,8 @@
 #include "vfile.h"
 #include "snprintf.h"
 #include "hooks.h"
+#include "common.h"
+#include "strutil.h"
 
 /*}}}*/
 
@@ -129,17 +131,20 @@ static void remove_group_entry (Slrn_Group_Type *);
 
 /*{{{ Functions that deal with Group Range */
 
-static int count_unread (Slrn_Range_Type *r)
+static NNTP_Artnum_Type count_unread (Slrn_Range_Type *r)
 {
-   int unread = r->max;
+   NNTP_Artnum_Type nread = 0;
+   NNTP_Artnum_Type rmax = r->max;
 
    while (r->next != NULL)
      {
 	r = r->next;
-	unread -= r->max - r->min + 1;
+	nread += r->max - r->min + 1;
      }
+   if (nread > rmax)
+     return 0;
    
-   return unread;
+   return rmax - nread;
 }
 
 void slrn_group_recount_unread (Slrn_Group_Type *g)
@@ -148,11 +153,9 @@ void slrn_group_recount_unread (Slrn_Group_Type *g)
    if (g->range.min>1)
      g->range.next = slrn_ranges_add (g->range.next, 1, g->range.min-1);
    g->unread = count_unread (&g->range);
-   if (g->unread < 0)
-     g->unread = 0;
 }   
 
-static int is_article_requested (Slrn_Group_Type *g, long num) /*{{{*/
+static int is_article_requested (Slrn_Group_Type *g, NNTP_Artnum_Type num) /*{{{*/
 {
 #if SLRN_HAS_SPOOL_SUPPORT
    if (Slrn_Server_Id != SLRN_SERVER_ID_SPOOL) return 0;
@@ -171,13 +174,13 @@ static int is_article_requested (Slrn_Group_Type *g, long num) /*{{{*/
 }
 /*}}}*/
 
-static void group_mark_article_as_read (Slrn_Group_Type *g, long num) /*{{{*/
+static void group_mark_article_as_read (Slrn_Group_Type *g, NNTP_Artnum_Type num) /*{{{*/
 {
    Slrn_Range_Type *r;
    
    /* Never mark articles as read if their body has been requested. */
    if (is_article_requested (g, num)) return;
-   
+
    r = &g->range;
    if (r->max < num)  /* not at server yet so update our data */
      {
@@ -195,54 +198,62 @@ static void group_mark_article_as_read (Slrn_Group_Type *g, long num) /*{{{*/
 	r = r->next;
      }
    
-   if (g->unread > 0) g->unread -= 1;
+   if (g->unread != 0) g->unread -= 1;
    Slrn_Groups_Dirty = 1;
    g->range.next = slrn_ranges_add (g->range.next, num, num);   
 }
 
 /*}}}*/
 
-void slrn_mark_article_as_read (char *group, long num) /*{{{*/
+void slrn_mark_articles_as_read (char *group, 
+				 NNTP_Artnum_Type rmin, NNTP_Artnum_Type rmax) /*{{{*/
 {
    Slrn_Group_Type *g;
-   unsigned long hash;
    
    if (group == NULL)
+     g = Slrn_Group_Current_Group;
+   else
      {
-	group_mark_article_as_read (Slrn_Group_Current_Group, num);
-	return;
+	unsigned long hash = slrn_compute_hash ((unsigned char *) group,
+						(unsigned char *) group + strlen (group));
+   
+	g = Group_Hash_Table[hash % GROUP_HASH_TABLE_SIZE];
+   
+	while (g != NULL)
+	  {
+	     if ((g->hash == hash) && !strcmp (group, g->group_name))
+	       {
+		  /* If it looks like we have read this group, mark it read. */
+		  if ((g->flags & GROUP_UNSUBSCRIBED)
+		      && (g->range.next == NULL))
+		    return;
+		  
+		  break;
+	       }
+	     g = g->hash_next;
+	  }
      }
    
-   hash = slrn_compute_hash ((unsigned char *) group,
-			     (unsigned char *) group + strlen (group));
-   
-   g = Group_Hash_Table[hash % GROUP_HASH_TABLE_SIZE];
-   
-   while (g != NULL)
+   if (g == NULL)
+     return;
+
+   while (rmin <= rmax)
      {
-	if ((g->hash == hash) && !strcmp (group, g->group_name))
-	  {
-	     /* If it looks like we have read this group, mark it read. */
-	     if (((g->flags & GROUP_UNSUBSCRIBED) == 0)
-		 || (g->range.next != NULL))
-	       group_mark_article_as_read (g, num);
-	     
-	     break;
-	  }
-	g = g->hash_next;
+	group_mark_article_as_read (g, rmin);
+	rmin++;
      }
 }
 
 /*}}}*/
 
-static int group_update_range (Slrn_Group_Type *g, int min, int max) /*{{{*/
+static int group_update_range (Slrn_Group_Type *g, NNTP_Artnum_Type min, NNTP_Artnum_Type max) /*{{{*/
 {
-   int n, max_available;
+   NNTP_Artnum_Type n, max_available;
    Slrn_Range_Type *r;
    
    if (max == 0)
      {
-	int nmax, nmin;
+	NNTP_Artnum_Type nmax, nmin;
 	
 	nmax = g->range.max;
 	nmin = g->range.min;
@@ -251,7 +262,7 @@ static int group_update_range (Slrn_Group_Type *g, int min, int max) /*{{{*/
 	if (nmax > 0)
 	  for (n = nmin; n <= nmax; n++)
 	    group_mark_article_as_read (g, n);
-	
+
 	/* g->unread = 0; */
 	Slrn_Full_Screen_Update = 1;
 	Slrn_Groups_Dirty = 1;
@@ -265,7 +276,7 @@ static int group_update_range (Slrn_Group_Type *g, int min, int max) /*{{{*/
     * mark any articles as read */
    if (Kill_After_Max && (max < g->range.max))
      {
-	int nmax = g->range.max;
+	NNTP_Artnum_Type nmax = g->range.max;
 	for (n = max + 1; n <= nmax; n++)
 	  group_mark_article_as_read (g, n);
      }
@@ -281,13 +292,6 @@ static int group_update_range (Slrn_Group_Type *g, int min, int max) /*{{{*/
    r = &g->range;
    if (r->next != NULL)
      {
-#if 0
-	/* stesch@parsec.rhein-neckar.de (Stefan Scholl) suggests moving this
-	 * below to avoid posibility of n < 0.
-	 */
-	if (r->next->min <= r->min)
-	  r->next->min = 1;
-#endif	
 	n = r->max;
 	while (r->next != NULL)
 	  {
@@ -297,11 +301,10 @@ static int group_update_range (Slrn_Group_Type *g, int min, int max) /*{{{*/
 	if (n < 0) n = 0;
 	if (n > max_available) n = max_available;
 	g->unread = n;
-#if 1
+
 	r = &g->range;
 	if (r->next->min <= r->min)
 	  r->next->min = 1;
-#endif
      }
    else
      g->unread = max_available;
@@ -319,7 +322,7 @@ static int group_update_range (Slrn_Group_Type *g, int min, int max) /*{{{*/
 
 /*}}}*/
 
-static int group_sync_group_with_server (Slrn_Group_Type *g, int *minp, int *maxp) /*{{{*/
+static int group_sync_group_with_server (Slrn_Group_Type *g, NNTP_Artnum_Type *minp, NNTP_Artnum_Type *maxp) /*{{{*/
 {
    char *group;
    int status, reselect;
@@ -340,8 +343,8 @@ static int group_sync_group_with_server (Slrn_Group_Type *g, int *minp, int *max
    
    if (status == ERR_NOGROUP)
      {
-	slrn_error (_("Group %s is bogus%s."), group,
-		    Slrn_Drop_Bogus_Groups ? _(" - dropping it") : "");
+	slrn_message_now (_("Group %s is bogus%s."), group,
+			  Slrn_Drop_Bogus_Groups ? _(" - dropping it") : "");
 	Slrn_Saw_Warning = 1;
 	if (Slrn_Drop_Bogus_Groups)
 	  remove_group_entry (g);
@@ -445,7 +448,7 @@ static Slrn_Group_Type *find_group_entry (char *name, unsigned int len) /*{{{*/
 
 /*}}}*/
 static Slrn_Group_Type *create_group_entry (char *name, unsigned int len, /*{{{*/
-					    int min, int max, int query_server,
+					    NNTP_Artnum_Type min, NNTP_Artnum_Type max, int query_server,
 					    int skip_find)
 {
    int hash_index;
@@ -473,7 +476,7 @@ static Slrn_Group_Type *create_group_entry (char *name, unsigned int len, /*{{{*
 	status = Slrn_Server_Obj->sv_select_group (g->group_name, &min, &max);
 	if (status == ERR_NOGROUP)
 	  {
-	     slrn_error (_("Group %s is bogus%s."), g->group_name,
+	     slrn_message_now (_("Group %s is bogus%s."), g->group_name,
 			 Slrn_Drop_Bogus_Groups ? _(" - ignoring it") : "");
 	     if (Slrn_Drop_Bogus_Groups)
 	       {
@@ -567,7 +570,8 @@ static int add_group (char *name, unsigned int len, /*{{{*/
 	     
 	     if (NULL != (tmp_name = slrn_strnmalloc (name, len, 1)))
 	       {
-		  slrn_error (_("Group %s is bogus - ignoring it."), tmp_name);
+		  slrn_message_now (_("Group %s is bogus - ignoring it."), tmp_name);
+		  Slrn_Saw_Warning = 1;
 		  slrn_free (tmp_name);
 	       }
 	     return -1;
@@ -776,15 +780,30 @@ static void toggle_group_formats (void)
 
 int slrn_group_search (char *str, int dir) /*{{{*/
 {
+#if SLANG_VERSION < 20000
    SLsearch_Type st;
+#else
+   SLsearch_Type *st = NULL;
+   unsigned int flags;
+#endif
    Slrn_Group_Type *g;
    int found = 0;
    
    g = Slrn_Group_Current_Group;
    if (g == NULL) return 0;
    
+#if SLANG_VERSION < 20000
    SLsearch_init (str, 1, 0, &st);
-   
+#else
+   flags = SLSEARCH_CASELESS;
+   if (Slrn_UTF8_Mode)
+     flags |= SLSEARCH_UTF8;
+
+   st = SLsearch_new ((SLuchar_Type *) str, flags);
+   if (st == NULL)
+     return 0;
+#endif
+
    do
      {
 	if (dir > 0)
@@ -801,6 +820,7 @@ int slrn_group_search (char *str, int dir) /*{{{*/
 	
 	if ((g->flags & GROUP_HIDDEN) == 0)
 	  {
+#if SLANG_VERSION < 20000
 	     if ((NULL != SLsearch ((unsigned char *) g->group_name,
 				    (unsigned char *) g->group_name + strlen (g->group_name),
 				   &st))
@@ -812,10 +832,25 @@ int slrn_group_search (char *str, int dir) /*{{{*/
 		  found = 1;
 		  break;
 	       }
+#else
+	     if ((NULL != SLsearch_forward (st, (unsigned char *) g->group_name,
+					   (unsigned char *) g->group_name + strlen (g->group_name)))
+		 || ((NULL != g->descript)
+		     && (NULL != SLsearch_forward (st, (unsigned char *) g->descript,
+						  (unsigned char *) g->descript + strlen (g->descript)))))
+	       {
+		  found = 1;
+		  break;
+	       }
+#endif
 	  }
      }
    while (g != Slrn_Group_Current_Group);
    
+#if SLANG_VERSION >= 20000
+   SLsearch_delete (st);
+#endif
+
    Slrn_Group_Current_Group = g;
    find_line_num ();
    return found;
@@ -841,7 +876,7 @@ unsigned int slrn_group_down_n (unsigned int n) /*{{{*/
 
 int slrn_group_select_group (void) /*{{{*/
 {
-   int min, max, n, max_available, last_n;
+   NNTP_Artnum_Type min, max, n, max_available, last_n;
    int ret;
    Slrn_Range_Type *r;
    int prefix;
@@ -875,19 +910,19 @@ int slrn_group_select_group (void) /*{{{*/
    if ((prefix != 0) || (n == 0))
      n = max_available;
 
-   if ((prefix & 1) ||
-       ((n > Slrn_Query_Group_Cutoff)
-	&& (Slrn_Query_Group_Cutoff > 0)) ||
-       ((n > -Slrn_Query_Group_Cutoff)
-	&& (Slrn_Query_Group_Cutoff < 0)))
+   if ((prefix & 1) 
+       || ((Slrn_Query_Group_Cutoff > 0) 
+	   && (n > (NNTP_Artnum_Type)Slrn_Query_Group_Cutoff))
+       || ((Slrn_Query_Group_Cutoff < 0) 
+	   && (n > (NNTP_Artnum_Type)(-Slrn_Query_Group_Cutoff))))
      {
-	char int_prompt_buf[256];
+	char int_prompt_buf[512];
 	if ((prefix & 1) || (Slrn_Query_Group_Cutoff > 0))
 	  {
 	     slrn_snprintf (int_prompt_buf, sizeof (int_prompt_buf),
 			    _("%s: Read how many? "),
 			    Slrn_Group_Current_Group->group_name);
-	     if ((-1 == slrn_read_integer (int_prompt_buf, &n, &n))
+	     if ((-1 == slrn_read_artnum_int (int_prompt_buf, &n, &n))
 		 || (n <= 0))
 	       {
 		  slrn_clear_message ();
@@ -897,7 +932,7 @@ int slrn_group_select_group (void) /*{{{*/
 	  }
 	else
 	  {
-	     slrn_message_now (_("Only downloading %d of %d articles."),
+	     slrn_message_now (_("Only downloading %d of " NNTP_FMT_ARTNUM " articles."),
 			       -Slrn_Query_Group_Cutoff, n);
 	     n = -Slrn_Query_Group_Cutoff;
 	  }
@@ -936,17 +971,17 @@ void slrn_select_next_group (void) /*{{{*/
    if (Slrn_Group_Current_Group == NULL) 
      return;
    
-   while ((SLang_Error == 0) && (1 == slrn_group_down_n (1)))
+   while ((SLang_get_error () == 0) && (1 == slrn_group_down_n (1)))
      {
 	if (Slrn_Group_Current_Group->unread == 0) 
 	  continue;
 	
 	if (0 == slrn_group_select_group ())
 	  break;
-	else if (SLang_Error == INTRINSIC_ERROR)
+	else if (SLang_get_error () == INTRINSIC_ERROR)
 	  /* all articles killed by scorefile, so proceed */
 	  {
-	     SLang_Error = 0;
+	     SLang_set_error (0);
 	     slrn_clear_message ();
 	  }
      }
@@ -958,7 +993,7 @@ void slrn_select_prev_group (void) /*{{{*/
    if (Slrn_Group_Current_Group == NULL)
      return;
 
-   while ((SLang_Error == 0) && (1 == slrn_group_up_n (1)))
+   while ((SLang_get_error () == 0) && (1 == slrn_group_up_n (1)))
      {
        if (Slrn_Group_Current_Group->unread == 0)
          continue;
@@ -1092,8 +1127,9 @@ static void refresh_groups (Slrn_Group_Type **c) /*{{{*/
 	       }
 	     if (ranges[j].min == -1)
 	       {
-		  slrn_error (_("Group %s is bogus%s."), g->group_name,
+		  slrn_message_now (_("Group %s is bogus%s."), g->group_name,
 			      Slrn_Drop_Bogus_Groups ? _(" - dropping it") : "");
+		  Slrn_Saw_Warning = 1;
 		  if (Slrn_Drop_Bogus_Groups)
 		    {
 		       Slrn_Group_Type *tmp = g->next;
@@ -1132,7 +1168,7 @@ static void refresh_groups_cmd (void) /*{{{*/
 	  /* hack: avoid a problem with inn not updating the high water mark */
 	  {
 	     Slrn_Group_Type *a = c, *b;
-	     int min, max;
+	     NNTP_Artnum_Type min, max;
 	     
 	     if (((NULL != (b = c->next)) || (NULL != (b = c->prev))) &&
 		 (0 == strcmp (c->group_name,
@@ -1141,7 +1177,7 @@ static void refresh_groups_cmd (void) /*{{{*/
 		  a = b;
 	       }
 	     
-	     group_sync_group_with_server (a, &min, &max);
+	     (void) group_sync_group_with_server (a, &min, &max);
 	  }
      }
    
@@ -1410,6 +1446,9 @@ static void subscribe (void) /*{{{*/
      }
    find_line_num ();
    Slrn_Full_Screen_Update = 1;
+#if SLANG_VERSION >= 20000
+   SLregexp_free (re);
+#endif
 }
 
 /*}}}*/
@@ -1479,6 +1518,9 @@ static void unsubscribe (void) /*{{{*/
      }
    find_line_num ();
    Slrn_Full_Screen_Update = 1;
+#if SLANG_VERSION >= 20000
+   SLregexp_free (re);
+#endif
 }
 
 /*}}}*/
@@ -1570,6 +1612,9 @@ static void toggle_list_all_groups1 (int hide_flag) /*{{{*/
 	       }
 	     g = g->next;
 	  }
+#if SLANG_VERSION >= 20000
+	SLregexp_free (re);
+#endif
      }
    
    g = Slrn_Group_Current_Group;
@@ -1693,16 +1738,14 @@ void slrn_post_cmd (void) /*{{{*/
 	return;
      }
 
-#if SLRN_HAS_SLANG
-   slrn_run_hooks (HOOK_POST, 0);
-   if (SLang_Error)
-     return;
-#endif
-
    if ((Slrn_User_Wants_Confirmation & SLRN_CONFIRM_POST) && (Slrn_Batch == 0) &&
        (slrn_get_yesno (1, _("Are you sure that you want to post")) <= 0))
      return;
    
+   slrn_run_hooks (HOOK_POST, 0);
+   if (SLang_get_error ())
+     return;
+   
    if (Slrn_Group_Current_Group == NULL)
      name = "";
    else name = Slrn_Group_Current_Group->group_name;
@@ -1711,7 +1754,7 @@ void slrn_post_cmd (void) /*{{{*/
      name = "";
    slrn_strncpy (group, name, sizeof (group));
    if (slrn_read_input (_("Newsgroup: "), NULL, group, 1, 1) <= 0) return;
-   if (strchr (group, ',') != NULL)
+   if (slrn_strbyte (group, ',') != NULL)
      {
 	slrn_strncpy (followupto, name, sizeof (followupto));
 	(void) slrn_read_input (_("Followup-To: "), NULL, followupto, 1, 1);
@@ -1799,9 +1842,7 @@ static void enter_group_mode_hook (void)
      Group_Window.cannot_scroll = 2;
    else
      Group_Window.cannot_scroll = SLtt_Term_Cannot_Scroll;
-#if SLRN_HAS_SLANG
    slrn_run_hooks (HOOK_GROUP_MODE, 0);
-#endif
 }
 
 static void group_winch_sig (int old_r, int old_c)
@@ -1974,7 +2015,7 @@ void slrn_init_group_mode (void) /*{{{*/
    char  *err = _("Unable to create group keymap!");
    
    if (NULL == (Slrn_Group_Keymap = SLang_create_keymap ("group", NULL)))
-     slrn_exit_error (err);
+     slrn_exit_error ("%s", err);
    
    Group_Mode_Cap.keymap = Slrn_Group_Keymap;
    
@@ -2062,7 +2103,7 @@ void slrn_init_group_mode (void) /*{{{*/
 #if USE_TEST_FUNCTION
    SLkm_define_key  ("y", (FVOID_STAR) test_function, Slrn_Group_Keymap);
 #endif
-   if (SLang_Error) slrn_exit_error (err);
+   if (SLang_get_error ()) slrn_exit_error ("%s", err);
 }
 
 /*}}}*/
@@ -2605,7 +2646,10 @@ static int read_and_parse_newsrc_file (void)
 	if ((p == pmax) || (p == vline))
 	  continue;
      
-	vline[vlen-1] = 0;	       /* kill \n and NULL terminate */
+	if (vline[vlen-1] == '\n')
+	  vline[vlen-1] = 0;
+	else
+	  vline[vlen] = 0;
 
 	if (-1 == add_group (vline, (unsigned int) (p - vline),
 			     ((ch == '!') ? GROUP_UNSUBSCRIBED : 0), 0, 0))
@@ -2802,7 +2846,7 @@ int slrn_write_newsrc (int auto_save) /*{{{*/
 	max = g->range.max;
 	if (r != NULL)
 	  {
-	     int max_newsrc_number=0;
+	     NNTP_Artnum_Type max_newsrc_number=0;
 	     /* Make this check because the unsubscribed group
 	      * range may not have been initialized from the server.
 	      */
@@ -2823,7 +2867,7 @@ int slrn_write_newsrc (int auto_save) /*{{{*/
 	  }
 	else if (g->range.min > 2)
 	  {
-	     if (fprintf (fp, " %d-%d", 1, g->range.min - 1) < 0)
+	     if (fprintf (fp, (" 1-" NNTP_FMT_ARTNUM), g->range.min - 1) < 0)
 	       goto write_error;
 	  }
 	
@@ -2880,7 +2924,7 @@ static void group_quick_help (void) /*{{{*/
    if (Slrn_Group_Help_Line != NULL)
      hlp = Slrn_Group_Help_Line;
    
-   if (0 == slrn_message (hlp))
+   if (0 == slrn_message ("%s", hlp))
      Slrn_Message_Present = 0;
 }
 
@@ -2889,7 +2933,7 @@ static void group_quick_help (void) /*{{{*/
 static char *group_display_format_cb (char ch, void *data, int *len, int *color) /*{{{*/
 {
    Slrn_Group_Type *g = (Slrn_Group_Type*) data;
-   static char buf[32];
+   static char buf[512];
    char *retval = buf;
    
    *retval = 0;
@@ -2918,7 +2962,7 @@ static char *group_display_format_cb (char ch, void *data, int *len, int *color)
 #ifdef HAVE_ANSI_SPRINTF
 	  *len =
 #endif
-	  sprintf (buf, "%d", g->range.max); /* safe */
+	  sprintf (buf, NNTP_FMT_ARTNUM, g->range.max); /* safe */
 	break;
       case 'l':
 	if (g->range.min == -1)
@@ -2930,7 +2974,7 @@ static char *group_display_format_cb (char ch, void *data, int *len, int *color)
 #ifdef HAVE_ANSI_SPRINTF
 	  *len =
 #endif
-	  sprintf (buf, "%d", g->range.min); /* safe */
+	  sprintf (buf, NNTP_FMT_ARTNUM, g->range.min); /* safe */
 	break;
       case 'n':
 	retval = g->group_name;
@@ -2946,13 +2990,13 @@ static char *group_display_format_cb (char ch, void *data, int *len, int *color)
 #ifdef HAVE_ANSI_SPRINTF
 	  *len =
 #endif
-	  sprintf (buf, "%d", g->range.max - g->range.min + 1); /* safe */
+	  sprintf (buf, NNTP_FMT_ARTNUM, g->range.max - g->range.min + 1); /* safe */
 	break;
       case 'u':
 #ifdef HAVE_ANSI_SPRINTF
 	*len =
 #endif
-	sprintf (buf, "%d", g->unread); /* safe */
+	sprintf (buf, NNTP_FMT_ARTNUM, g->unread); /* safe */
 	break;
      }
    
@@ -3072,12 +3116,8 @@ static void group_update_screen (void) /*{{{*/
 /*}}}*/
 
 /* intrinsic functions */
-#if SLRN_HAS_SLANG
 void slrn_intr_get_group_order (void) /*{{{*/
 {
-# if SLANG_VERSION < 10400
-   slrn_error (_("To use this feature, please update s-lang and recompile slrn."));
-# else
    Slrn_Group_Type *g;
    SLang_Array_Type *retval;
    int n = 0;
@@ -3107,17 +3147,13 @@ void slrn_intr_get_group_order (void) /*{{{*/
      }
    
    (void) SLang_push_array (retval, 1);
-# endif /* SLANG_VERSION */
 }
 /*}}}*/
 
 void slrn_intr_set_group_order (void) /*{{{*/
 {
-#if SLANG_VERSION < 10400
-   slrn_error (_("To use this feature, please update s-lang and recompile slrn."));
-# else
    SLang_Array_Type *at;
-   Slrn_Group_Type *last = NULL, *rest = Groups;
+   Slrn_Group_Type *last, *rest, *g;
    int i, rows;
    
    if (-1 == SLang_pop_array_of_type (&at, SLANG_STRING_TYPE))
@@ -3132,44 +3168,69 @@ void slrn_intr_set_group_order (void) /*{{{*/
 	SLang_free_array (at);
 	return;
      }
-   
+
+   if ((Groups == NULL) || (Groups->next == NULL))
+     {
+	SLang_free_array (at);
+	return;
+     }
+
    rows = at->dims[0];
-   
+   last = NULL;
+   rest = Groups;
+
    for (i = 0; i < rows; i++)
      {
 	char *name;
-	Slrn_Group_Type *g;
-	
-	(void) SLang_get_array_element (at, &i, &name);
 	
+	if ((-1 == SLang_get_array_element (at, &i, &name))
+	    || (name == NULL))
+	  continue;
+
 	if (NULL == (g = find_group_entry (name, strlen (name))))
 	  continue;
-	
+
+	/* It is possible for name to occur multiple times in the array.
+	 * So we have to be careful.  There are two lists here: the part
+	 * that runs from Groups to last, and the rest. The first group 
+	 * is the already sorted part, and the second is unsorted. 
+	 * Ordinarily g will come from the second (rest), but if name appears
+	 * more than once in the array, it will come from the first part when
+	 * seen the second time.
+	 */
+	if (g == last)		       /* at tail of first part */
+	  continue;
+
+	if (g == Groups)	       /* at head of first part */
+	  Groups = g->next;
+
+	if (g == rest)		       /* at head of second part */
+	  rest = rest->next;
+
+	/* Remove it from its current location */
 	if (g->next != NULL)
 	  g->next->prev = g->prev;
 	if (g->prev != NULL)
 	  g->prev->next = g->next;
-	g->prev = last;
-	if (last == NULL)
+
+	g->prev = last;		       /* append it to first part */
+	
+	if (last == NULL)	       /* the head of the first part */
 	  Groups = g;
 	else
 	  last->next = g;
-	if (g == rest)
-	  rest = rest->next;
+
+	g->next = rest;		       /* connect up the rest */
+	if (rest != NULL)
+	  rest->prev = g;
+
 	last = g;
      }
-   
-   if (last != NULL)
-     last->next = rest;
-   if (rest != NULL)
-     rest->prev = last;
-   
+
    SLang_free_array (at);
-   
+
    find_line_num ();
    Slrn_Full_Screen_Update = 1;
    Slrn_Groups_Dirty = 1;
-# endif /* SLANG_VERSION */
 }
 /*}}}*/
-#endif /* SLRN_HAS_SLANG */
