@@ -1,8 +1,8 @@
 /*
  This file is part of SLRN.
 
- Copyright (c) 1994, 1999 John E. Davis <davis@space.mit.edu>
- Copyright (c) 2001 Thomas Schultz <tststs@gmx.de>
+ Copyright (c) 1994, 1999, 2007-2008 John E. Davis <jed@jedsoft.org>
+ Copyright (c) 2001-2006 Thomas Schultz <tststs@gmx.de>
 
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
@@ -73,7 +73,7 @@ static int _nntp_interrupt_handler (void)
 {
    (void) slrn_handle_interrupts ();
 
-   if ((SLang_Error == USER_BREAK) || SLKeyBoard_Quit)
+   if ((SLang_get_error () == USER_BREAK) || SLKeyBoard_Quit)
      return -1;
 
    return 0;
@@ -108,21 +108,45 @@ static int _nntp_initialize_server (void)
    if (NULL == (NNTP_Server = nntp_open_server (NNTP_Server_Name, NNTP_Port)))
      return -1;
    
-   if (Slrn_Prefer_Head == 2)
+   NNTP_Server_Obj.sv_has_xover = 0;
+   if (Slrn_Prefer_Head & 2)
      NNTP_Server->can_xover = 0;
-   else if (-1 == nntp_has_cmd (NNTP_Server, "XOVER"))
+   else 
      {
-	_nntp_close_server ();
-	return -1;
+	if (-1 == nntp_has_cmd (NNTP_Server, "XOVER"))
+	  {
+	     _nntp_close_server ();
+	     return -1;
+	  }
+
+	if (NNTP_Server->can_xover == 0)
+	  {
+	     slrn_message_now (_("Server %s does not implement the XOVER command."), 
+			       NNTP_Server->host);
+	  }
+	else NNTP_Server_Obj.sv_has_xover = 1;
      }
-   
-   if (NNTP_Server->can_xover == 0)
+
+   NNTP_Server_Obj.sv_has_xhdr = 0;
+   if (Slrn_Prefer_Head & 1)
+     NNTP_Server->can_xhdr = 0;
+   else 
      {
-	slrn_message_now (_("Server %s does not implement the XOVER command."), 
-			  NNTP_Server->host);
-	NNTP_Server_Obj.sv_has_xover = 0;
+	if (-1 == nntp_has_cmd (NNTP_Server, "XHDR Path"))
+	  {
+	     _nntp_close_server ();
+	     return -1;
+	  }
+
+	if (NNTP_Server->can_xhdr == 0)
+	  {
+	     slrn_message_now (_("Server %s does not implement the XHDR command."),
+			       NNTP_Server->host);
+	  }
+	else NNTP_Server_Obj.sv_has_xhdr = 1;
      }
-   else NNTP_Server_Obj.sv_has_xover = 1;
+
+   NNTP_Server_Obj.sv_id = NNTP_Server->sv_id;
    
    NNTP_Post_Obj.po_can_post = NNTP_Server->can_post;
    NNTP_Server->flags |= NNTP_RECONNECT_OK;
@@ -152,17 +176,24 @@ static int _nntp_read_line (char *buf, unsigned int len)
    return -1;
 }
 
-static int _nntp_head_cmd (int id, char *msgid, int *real_idp)
+static int _nntp_get_article_size (NNTP_Artnum_Type id)
+{
+   /* no nntp-command for this, reading the article takes too much time */
+   (void) id;
+   return 0;
+}
+
+static int _nntp_head_cmd (NNTP_Artnum_Type id, char *msgid, NNTP_Artnum_Type *real_idp)
 {
    return nntp_head_cmd (NNTP_Server, id, msgid, real_idp);
 }
 
-static int _nntp_select_article (int n, char *msgid)
+static int _nntp_select_article (NNTP_Artnum_Type n, char *msgid)
 {
    return nntp_article_cmd (NNTP_Server, n, msgid);
 }
 
-static int _nntp_select_group (char *grp, int *min, int *max)
+static int _nntp_select_group (char *grp, NNTP_Artnum_Type *min, NNTP_Artnum_Type *max)
 {
    return nntp_select_group (NNTP_Server, grp, min, max);
 }
@@ -189,12 +220,12 @@ static int _nntp_put_server_cmd (char *cmd, char *buf, unsigned int len)
    return code;
 }
 
-static int _nntp_xpat_cmd (char *hdr, int rmin, int rmax, char *pat)
+static int _nntp_xpat_cmd (char *hdr, NNTP_Artnum_Type rmin, NNTP_Artnum_Type rmax, char *pat)
 {
    return nntp_xpat_cmd (NNTP_Server, hdr, rmin, rmax, pat);
 }
 
-static int _nntp_one_xhdr_cmd (char *hdr, int num, 
+static int _nntp_one_xhdr_cmd (char *hdr, NNTP_Artnum_Type num,
 			       char *buf, unsigned int buflen)
 {
    return nntp_one_xhdr_cmd (NNTP_Server, hdr, num, buf, buflen);
@@ -229,7 +260,7 @@ static int _nntp_list_active (char *pat)
 
 static int _nntp_send_authinfo (void)
 {
-   return nntp_authorization (NNTP_Server);
+   return nntp_authorization (NNTP_Server, Slrn_Force_Authentication);
 }
 
 static int _nntp_start_post (void)
@@ -276,7 +307,7 @@ static int _nntp_po_puts (char *buf)
    
    /* make sure \n --> \r\n */
    b = buf;
-   while (NULL != (b = slrn_strchr (buf, '\n')))
+   while (NULL != (b = slrn_strbyte (buf, '\n')))
      {
 	unsigned int len;
 	
@@ -291,22 +322,31 @@ static int _nntp_po_puts (char *buf)
    return nntp_fputs_server (NNTP_Server, buf);
 }
 
-static int _nntp_po_printf (char *fmt, ...)
+static int _nntp_po_vprintf (const char *fmt, va_list ap)
 {
-   va_list ap;
    char buf[NNTP_BUFFER_SIZE];
-   
-   va_start (ap, fmt);
+
    slrn_vsnprintf (buf, sizeof (buf), fmt, ap);
-   va_end (ap);
-   
+
    return _nntp_po_puts (buf);
 }
 
-static int _nntp_xover_cmd (int min, int max)
-{   
+static int _nntp_po_printf (char *fmt, ...)
+{
+   va_list ap;
+   int retval;
+
+   va_start(ap, fmt);
+   retval = _nntp_po_vprintf (fmt, ap);
+   va_end(ap);
+
+   return retval;
+}
+
+static int _nntp_xover_cmd (NNTP_Artnum_Type min, NNTP_Artnum_Type max)
+{
    int status;
-   
+
    if (OK_XOVER == (status = nntp_xover_cmd (NNTP_Server, min, max)))
      {
 	_NNTP_Abort_On_Disconnection = -1;
@@ -315,7 +355,7 @@ static int _nntp_xover_cmd (int min, int max)
    return status;
 }
 
-static int _nntp_xhdr_cmd (char *field, int min, int max)
+static int _nntp_xhdr_cmd (char *field, NNTP_Artnum_Type min, NNTP_Artnum_Type max)
 {
    int status;
    
@@ -327,7 +367,7 @@ static int _nntp_xhdr_cmd (char *field, int min, int max)
    return status;
 }
 
-static int _nntp_next_cmd (int *id)
+static int _nntp_next_cmd (NNTP_Artnum_Type *id)
 {
    return nntp_next_cmd (NNTP_Server, id);
 }
@@ -343,12 +383,47 @@ static unsigned int _nntp_get_bytes (int clear)
    return temp;
 }
 
+static char * _nntp_get_recom_id (void)
+{
+   SLRegexp_Type *re;
+   
+   if ((re=slrn_compile_regexp_pattern("<.*@.*\\..*>")) != NULL)
+     {
+	char *t, *msgid = NULL, *post_rsp;
+	post_rsp = NNTP_Server->rspbuf;
+#if SLANG_VERSION < 20000
+	t=(char*)SLang_regexp_match((unsigned char*) post_rsp,
+				    strlen(post_rsp), re);
+#else
+	t = SLregexp_match (re, post_rsp, strlen (post_rsp));
+#endif
+	if (t != NULL)
+	  {
+	     unsigned int len;
+#if SLANG_VERSION < 20000
+	     len = re->end_matches[0];
+#else
+	     (void) SLregexp_nth_match (re, 0, NULL, &len);
+#endif
+	     
+	     msgid=slrn_strnmalloc(t, len, 1);
+	  }
+#if SLANG_VERSION >= 20000
+	SLregexp_free (re);
+#endif
+	return msgid;
+     }
+   return NULL;
+}
+
 static int nntp_init_objects (void)
 {
    NNTP_Post_Obj.po_start = _nntp_start_post;
    NNTP_Post_Obj.po_end = _nntp_end_post;
    NNTP_Post_Obj.po_printf = _nntp_po_printf;
+   NNTP_Post_Obj.po_vprintf = _nntp_po_vprintf;
    NNTP_Post_Obj.po_puts = _nntp_po_puts;
+   NNTP_Post_Obj.po_get_recom_id = _nntp_get_recom_id;
    NNTP_Post_Obj.po_can_post = 1;
    
    NNTP_Server_Obj.sv_select_group = _nntp_select_group;
@@ -356,6 +431,7 @@ static int nntp_init_objects (void)
    NNTP_Server_Obj.sv_current_group = _nntp_current_group;
    NNTP_Server_Obj.sv_read_line = _nntp_read_line;
    NNTP_Server_Obj.sv_close = _nntp_close_server;
+   NNTP_Server_Obj.sv_get_article_size = _nntp_get_article_size;
    NNTP_Server_Obj.sv_initialize = _nntp_initialize_server;
    NNTP_Server_Obj.sv_select_article = _nntp_select_article;
    NNTP_Server_Obj.sv_put_server_cmd = _nntp_put_server_cmd;
@@ -375,6 +451,7 @@ static int nntp_init_objects (void)
    NNTP_Server_Obj.sv_nntp_next = _nntp_next_cmd;
    NNTP_Server_Obj.sv_reset = _nntp_reset;
    NNTP_Server_Obj.sv_nntp_bytes = _nntp_get_bytes;
+   NNTP_Server_Obj.sv_id = SERVER_ID_UNKNOWN;
    return 0;
 }
 
